% This script compares the timings for LAPLACE_MERGE and LAPLACE_RECURSIVE.
% It reproduces results shown in the technical report
% Recursive blocked algorithms for linear systems with Kronecker product structure

d = 5;

if d == 3,
    
nmax = 200;
nmax_rec = 200;
nmax_sylv = 125;
nvec = 5:5:nmax;

time_rec = []; time_merge = []; time_sylv = []; 
for nn = nvec,
    nn,
    n = nn*ones(1,d);
    for i = 1:d, A{i} = randn(n(i)); end
    B = randn(n);
    
    tt = 0; s = 0;
    while tt < 1,
        s = s + 1;
        tic
        X = laplace_merge(A,B);
        tt = tt + toc;
    end
    time_merge = [time_merge;tt/s];
    time_merge(end)
    if laplace_residual(A,B,X) > 10^(-3),
        warning('Problem');
    end

    if nn <= nmax_rec
    
    tt = 0; s = 0;
    while tt < 1,
        s = s + 1;
        tic
        X = laplace_recursive(A,B);
        tt = tt + toc;
    end
    time_rec = [time_rec;tt/s];
    time_rec(end)
    
    if laplace_residual(A,B,X) > 10^(-3),
        warning('Problem');
    end
    end
    
    if nn <= nmax_sylv,
    
    tt = 0; s = 0;
    while tt < 1,
        s = s + 1;
        tic
    
    C = B;
    for i = 1:d,
       [Q{i},R{i}] = schur(A{i},'complex');
       order = [i, 1:i-1, i+1:d];
       C = permute(C,order);
       C = Q{i}'*reshape( C, n(i), prod( n( order(2:d) ) ) );
       C = reshape( C, n(order) ); C = ipermute(C,order);
    end

    AA = kroneckerize(R{1},R{2});
    BB = reshape( C, n(1)*n(2), n(3) );
    % Pre-Matlab 2018
    % XX = sylvslv('C', AA, R{3}.', BB); 
    XX = matlab.internal.math.sylvester_tri( AA, conj( R{3} ), BB, 'I', 'I', 'trans');

    
    X = C;
    X(:) = XX;
    
    for i = 1:d,
       order = [i, 1:i-1, i+1:d];
       X = permute(X,order);
       X = Q{i}*reshape( X, n(i), prod( n( order(2:d) ) ) );
       X = reshape( X, n(order) ); X = ipermute(X,order);
    end
        tt = tt + toc;
    end
    
    time_sylv = [time_sylv;tt/s];    
    time_sylv(end)
    
    if laplace_residual(A,B,X) > 10^(-3),
        warning('Problem');
    end

    end
    
end

end


if d == 4,
    
nmax = 80;
nmax_sylv = 80;
nmax_rec = 55;
nvec = 5:5:nmax;

time_rec = []; time_sylv = []; time_merge = [];
for nn = nvec,
    nn,
    n = nn*ones(1,d);
    for i = 1:d, A{i} = randn(n(i)); end
    B = randn(n);
    
    tt = 0; s = 0;
    while tt < 1,
        s = s + 1;
        tic
        X = laplace_merge(A,B);
        tt = tt + toc;
    end
    time_merge = [time_merge;tt/s];
    time_merge(end)
    if laplace_residual(A,B,X) > 10^(-3),
        warning('Problem');
    end

    if nn <= nmax_rec
    tt = 0; s = 0;
    while tt < 1,
        s = s + 1;
        tic
        X = laplace_recursive(A,B);
        tt = tt + toc;
    end
    time_rec = [time_rec;tt/s];
    time_rec(end)

    if laplace_residual(A,B,X) > 10^(-3),
        warning('Problem');
    end
    end
    
    
    if nn <= nmax_sylv,
    tt = 0; s = 0;
    while tt < 1,
        s = s + 1;
        tic
    
    C = B;
    for i = 1:d,
       [Q{i},R{i}] = schur(A{i},'complex');
       order = [i, 1:i-1, i+1:d];
       C = permute(C,order);
       C = Q{i}'*reshape( C, n(i), prod( n( order(2:d) ) ) );
       C = reshape( C, n(order) ); C = ipermute(C,order);
    end

    AA1 = kroneckerize(R{1},R{2});
    AA2 = kroneckerize(R{3},R{4});
    BB = reshape( C, n(1)*n(2), n(3)*n(4) );
    % Pre-Matlab 2018
    % XX = sylvslv('C', AA1, AA2.', BB); 
    XX = matlab.internal.math.sylvester_tri( AA1, conj( AA2 ), BB, 'I', 'I', 'trans');
    X = B;
    X(:) = XX;
    
    for i = 1:d,
       order = [i, 1:i-1, i+1:d];
       X = permute(X,order);
       X = Q{i}*reshape( X, n(i), prod( n( order(2:d) ) ) );
       X = reshape( X, n(order) ); X = ipermute(X,order);
    end
    if laplace_residual(A,B,X) > 10^(-3),
        warning('Problem');
    end
            tt = tt + toc;
    end


    time_sylv = [time_sylv;tt/s];
    time_sylv(end)
    end
    
end

end

if d == 5,
    
nmax = 30;
nmax_sylv = 25;
nvec = 5:5:nmax;

time_sylv = []; time_merge = [];
for nn = nvec,
    nn,
    n = nn*ones(1,d);
    for i = 1:d, A{i} = randn(n(i)); end
    B = randn(n);
    
    tt = 0; s = 0;
    while tt < 1,
        s = s + 1;
        tic
        X = laplace_merge(A,B);
        tt = tt + toc;
    end
    time_merge = [time_merge;tt/s];
    time_merge(end)
    if laplace_residual(A,B,X) > 10^(-3),
        warning('Problem');
    end    
    
    if nn <= nmax_sylv,
    tt = 0; s = 0;
    while tt < 1,
        s = s + 1;
        tic
    
    C = B;
    for i = 1:d,
       [Q{i},R{i}] = schur(A{i},'complex');
       order = [i, 1:i-1, i+1:d];
       C = permute(C,order);
       C = Q{i}'*reshape( C, n(i), prod( n( order(2:d) ) ) );
       C = reshape( C, n(order) ); C = ipermute(C,order);
    end

    AA1 = kroneckerize(R{1},R{2});
    AA2 = kroneckerize( kroneckerize(R{3},R{4}), R{5} );
    BB = reshape( C, n(1)*n(2), n(3)*n(4)*n(5) );
    % Pre-Matlab 2018
    % XX = sylvslv('C', AA1, AA2.', BB); 
    XX = matlab.internal.math.sylvester_tri( AA1, conj( AA2 ), BB, 'I', 'I', 'trans');
    X = B;
    X(:) = XX;
    
    for i = 1:d,
       order = [i, 1:i-1, i+1:d];
       X = permute(X,order);
       X = Q{i}*reshape( X, n(i), prod( n( order(2:d) ) ) );
       X = reshape( X, n(order) ); X = ipermute(X,order);
    end
    if laplace_residual(A,B,X) > 10^(-3),
        warning('Problem');
    end
            tt = tt + toc;
    end


    time_sylv = [time_sylv;tt/s];
    time_sylv(end)
    end
    
end

end

if d < 5,
semilogy(nvec(1:length(time_sylv)), time_sylv, 'k+-')
hold on
semilogy(nvec(1:length(time_rec)), time_rec, 'b*-')
semilogy(nvec(1:length(time_merge)), time_merge, 'rx-')
legend('Sylvester solver', 'Algorithm 2', 'Algorithm 3', 'Location', 'SouthEast')
hold off
else
semilogy(nvec(1:length(time_sylv)), time_sylv, 'k+-')
hold on
semilogy(nvec(1:length(time_merge)), time_merge, 'rx-')
legend('Sylvester solver',  'Algorithm 3', 'Location', 'SouthEast')
hold off    
end

