function X = gsylv_recursive( A, B, C, nmin, upflag )
% GSYLV_RECURSIVE( A, B, C ) computes the solution of the generalized Sylvester
% equation with Kronecker structure:
%   X x_1 A{1} + X x_1 C x_2 A{2} ... x_d A{d} = B.
% GSYLV_RECURSIVE( A, B, C, nmin ) allows to specify the minimal block
% size.

n=size(B); d=length(n);

if nargin < 4
    if d == 2
        nmin = 15;
    elseif d == 3
        nmin = 8;
    elseif d == 4
        nmin = 6;
    else
        nmin = 4;
    end
end

if nargin < 5
    upflag = 0;
end

if ~upflag,
   [A{1}, C, Q{1}, Z] = qz( A{1}, C, 'complex');
   B = Q{1}*reshape( B, n(1), prod( n( 2:end) ) );
   B = reshape( B, n );
   for i=2:d
       [Q{i}, A{i}] = schur( A{i}, 'complex' );
       order = [i, 1:i-1, i+1:d];
       B = permute( B, order );
       B = Q{i}'*reshape( B, n(i), prod( n( order(2:d) ) ) );
       B = reshape( B, n(order) ); B = ipermute( B, order );
   end
end

X = ten_triu( A, B, C, nmin );

if ~upflag
   X = Z*reshape( X, n(1), prod( n( 2:end ) ) );
   X = reshape( X, n );

   for i = 2:d
      order = [i, 1:i-1, i+1:d];
      X = permute(X,order);
      X = Q{i}*reshape( X, n(i), prod( n( order(2:d) ) ) );
      X = reshape( X, n(order) ); X = ipermute(X,order);
   end
end
end

function X = ten_triu( A, B, C, nmin )
n = size(B); d = length(n);

if max(n) <= nmin
    X = gsylv_small( A, B, C );
    return
end

[~,i] = max(n);

k=round(n(i)/2); if A{i}(k+1,k), k=k-1;end

Arec = A; Arec{i} = Arec{i}(k+1:n(i),k+1:n(i));
X = zeros(n);

for j=1:d, ind2{j}=':'; end, ind2{i}=k+1:n(i);
ind1=ind2; ind1{i}=1:k;

if i==1
    X2 = ten_triu(Arec, B(ind2{:}), C(k+1:n(1),k+1:n(1)), nmin );
else
    X2 = ten_triu(Arec, B(ind2{:}), C, nmin );
end

X(ind2{:}) = X2;
order = [i, 1:i-1, i+1:d];
X2 = permute(X2,order);

if i==1
    U1 = A{1}(1:k,k+1:n(1)) * reshape( X2, n(1)-k, prod( n(2:d ) ) );
    U1 = reshape( U1, [ k, n(2:d) ] ); 
    X2 = C(1:k,k+1:n(1)) * reshape( X2, n(1)-k, prod( n( 2:d ) ) );
    X2 = reshape( X2, [ k, n(2:d) ] ); 

    n2=size(X2);
    for j=2:d
        order = [j, 1:j-1, j+1:d];
        X2 = permute( X2, order );
        X2 = A{j}*reshape( X2, n2(j), prod( n2( order(2:d) ) ) );
        X2 = reshape( X2,  n2(order)  ); 
        X2 = ipermute(X2,order);
    end    
else
    X2 = A{i}(1:k,k+1:n(i)) * reshape( X2, n(i)-k, prod( n( order(2:d) ) ) );
    X2 = reshape( X2, [ k, n(order(2:d)) ] ); 
    X2 = ipermute( X2, order ); 

    n2 = size(X2);
    
    X2 = C*reshape(X2,n(1),prod(n2(2:end)));
    X2 = reshape(X2,n2);

    for j=2:i-1
        order = [j, 1:j-1, j+1:d];
        X2 = permute(X2,order);
        X2 = A{j} * reshape( X2, n(j), prod( n2( order(2:d) ) ) );
        X2 = reshape( X2,  n2(order)  ); 
        X2 = ipermute(X2,order);
    end

    for j=i+1:d
        order = [j, 1:j-1, j+1:d];
        X2 = permute(X2,order);
        X2 = A{j}* reshape( X2, n(j), prod( n2( order(2:d) ) ) );
        X2 = reshape( X2,  n2(order)  ); 
        X2 = ipermute(X2,order);
    end

end

Arec{i} = A{i}(1:k,1:k);
if i==1
    X1 = ten_triu( Arec, B(ind1{:})-U1-X2, C(1:k,1:k), nmin );
else
    X1 = ten_triu( Arec, B(ind1{:})-X2, C, nmin );
end
X(ind1{:}) = X1;

end
