function X = gsylv_small( A, B, C )
% GSYLV_SMALL( A, B, C ) computes the solution of the generalized Sylvester
% equation with Kronecker structure:
%   X x_1 A{1} + X x_1 C x_2 A{2} ... x_d A{d} = B.
% The usefulness of this function is restricted to small orders/sizes.

n = size(B); d = length(n);
if prod(n)>10000
    error('matrix too large')
end

K = C;
for i = 2:d
    K = kron(A{i},K);
end

nn = prod(n(2:end));
for i = 1:nn
        ik = 1+(i-1)*n(1):i*n(1);
        K(ik,ik) = K(ik,ik) + A{1};
end
X = B;
X(:) = K \ B(:);
