function X = laplace_recursive( A, B, nmin )
% LAPLACE_RECURSIVE( A, B ) computes the solution of the Laplace equation
%   X x_1 A{1} + ... + X x_d A{d} = B.
% LAPLACE_RECURSIVE( A, B, nmin ) allows to specify the minimal block
% size.

n = size(B); d = length(n);

if nargin < 3,
    if d == 3,
        nmin = 7;
    else
        nmin = 3;
    end
end

for i = 1:d,
    [Q{i},A{i}] = schur( A{i}, 'complex' );
    order = [i, 1:i-1, i+1:d];
    B = permute(B,order);
    B = Q{i}'*reshape( B, n(i), prod( n( order(2:d) ) ) );
    B = reshape( B, n(order) ); B = ipermute(B,order);
end

X = laplace_tri( A, B, nmin );

for i = 1:d,
    order = [i, 1:i-1, i+1:d];
    X = permute(X,order);
    X = Q{i}*reshape( X, n(i), prod( n( order(2:d) ) ) );
    X = reshape( X, n(order) ); X = ipermute(X,order);
end

end

function X = laplace_tri(A, B, nmin )
    n = size(B); d = length(n);
       
    if max(n) <= nmin,
        X = laplace_small( A, B );
        return
    end
    [~,i] = max(n);
    
    k = floor( n(i) / 2 ); if A{i}(k+1,k), k = k + 1; end
    
    Arec = A; Arec{i} = Arec{i}(k+1:n(i),k+1:n(i));
    B = double(B);
    X = zeros(size(B));
    
    for j=1:d, ind2{j}=':'; end, ind2{i}=k+1:n(i); ind1=ind2; ind1{i}=1:k;
    X2 = laplace_tri( Arec, B(ind2{:}), nmin );
    X(ind2{:})=X2;
    
    order = [i, 1:i-1, i+1:d]; 
    X2 = permute(X2,order);
    B1 = A{i}(1:k,k+1:n(i)) * reshape( X2, n(i)-k, prod( n( order(2:d) ) ) );
    B1 = reshape( B1, [ k, n(order(2:d)) ] );
    B1 = ipermute(B1,order);
    B(ind1{:})=B(ind1{:})-B1;
    
    Arec = A; Arec{i} = Arec{i}(1:k,1:k);

    X(ind1{:}) = laplace_tri( Arec, B(ind1{:}), nmin );
end